@@ -1,7 +1,7 @@
 /*
  * iucode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
  *
- * Copyright (c) 2010-2017 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
+ * Copyright (c) 2010-2018 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
  *               2000 Simon Trimmer, Tigran Aivazian
  *
  * Some code copied from microcode.ctl v1.17.
@@ -177,12 +177,33 @@ struct microcode_iterator_data {
 static struct microcode_iterator_data microcode_iterator_data;
 
 /* Filter masks */
+#define IUCODE_FILTERMASK_SCANCPUS 0xffffffffU
+
+enum iuc_rev_match_mode {
+	IUCODE_REVFLT_ANY = 0,	/* Ignore revision on match */
+	IUCODE_REVFLT_EQ,	/* Revision must be = filter's rev */
+	IUCODE_REVFLT_LT,   	/* Revision must be < filter's rev */
+	IUCODE_REVFLT_GT,	/* Revision must be > filter's rev */
+
+	IUCODE_REVFLT_SIZE	/* EOL */
+};
+static const char * const iuc_rev_match_mode_s[IUCODE_REVFLT_SIZE] = {
+	[IUCODE_REVFLT_ANY] = NULL,
+	[IUCODE_REVFLT_EQ] = "eq:",
+	[IUCODE_REVFLT_LT] = "lt:",
+	[IUCODE_REVFLT_GT] = "gt:",
+}; /* keep this in sync with enum iuc_rev_match_mode !! */
+
 struct microcode_filter_entry {
 	struct microcode_filter_entry *next;
 	uint32_t cpuid;		/* exact match */
 	uint32_t pfm;		/* common bits set match */
-	int invert;
+	int32_t  rev;
+	enum iuc_rev_match_mode rev_match;
+	int      invert;
 };
+static struct microcode_filter_entry *uc_filter_queue = NULL;
+static struct microcode_filter_entry *uc_filter_queue_tail = NULL;
 static struct microcode_filter_entry *uc_filter_list = NULL;
 static int filter_list_allow = 1;
 
@@ -247,6 +268,31 @@ static int parse_u32(const char *nptr, char **endptr, int base,
 	return 0;
 }
 
+static int parse_s32e(const char *nptr, char **endptr, int base,
+		      int32_t * const res)
+{
+	long int l;
+
+	assert(nptr);
+	assert(endptr);
+	assert(res);
+
+	errno = 0;
+	l = strtol(nptr, endptr, base);
+	if (errno || nptr == *endptr)
+		return errno ? errno : EINVAL;
+	if (l > UINT32_MAX || l < INT32_MIN)
+		return ERANGE;
+
+	/*
+	 * Accept UINT32_MAX >= x > INT32_MAX, as if it were
+	 * the binary representation of int32_t read as uint32_t
+	 */
+	*res = (int32_t)(l & UINT32_MAX);
+
+	return 0;
+}
+
 static int is_valid_fd(const int fd)
 {
 	return fcntl(fd, F_GETFD) != -1 || errno != EBADF;
@@ -1898,7 +1944,6 @@ static int is_in_date_range(const struct intel_ucode_metadata * const m)
 	return !!((d > datefilter_min) && (d < datefilter_max));
 }
 
-static void free_filter_list(struct microcode_filter_entry *f) __attribute__((unused));
 static void free_filter_list(struct microcode_filter_entry *f)
 {
 	static struct microcode_filter_entry *p;
@@ -1910,9 +1955,15 @@ static void free_filter_list(struct microcode_filter_entry *f)
 	}
 }
 
-static void xx_merge_filter(struct microcode_filter_entry * const f,
-			 const uint32_t pf_mask, const int invert)
+static int xx_merge_filter(struct microcode_filter_entry * const f,
+			 const uint32_t pf_mask, const int32_t rev,
+			 const enum iuc_rev_match_mode rev_match,
+			 const int invert)
 {
+	/* FIXME: we could do better, sometimes it is possible to merge */
+	if (f->rev_match != rev_match || f->rev != rev)
+		return 0; /* can't merge */
+
 	if (f->invert == invert) {
 		f->pfm |= pf_mask;
 	} else if (!(f->invert)) {
@@ -1921,15 +1972,21 @@ static void xx_merge_filter(struct microcode_filter_entry * const f,
 		f->invert = 0;
 		f->pfm = pf_mask;
 	}
+
+	return 1; /* merged */
 }
 
-static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int invert,
-				struct microcode_filter_entry ** const base)
+static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int32_t rev,
+			      enum iuc_rev_match_mode rev_match, int invert,
+			      struct microcode_filter_entry ** const base)
 {
 	struct microcode_filter_entry **pnext, *n;
 
 	assert(base);
 
+	if (!cpuid || cpuid == IUCODE_FILTERMASK_SCANCPUS)
+		return EINVAL;
+
 	if (!pf_mask)
 		pf_mask = PFMASK_MATCH_ANY;
 
@@ -1940,13 +1997,16 @@ static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int invert,
 
 	if (*pnext && (*pnext)->cpuid == cpuid) {
 		if (((pf_mask & (*pnext)->pfm) == pf_mask) &&
-		     ((*pnext)->invert == invert)) {
+		    (*pnext)->rev_match == rev_match &&
+		    (*pnext)->rev == rev &&
+		    (*pnext)->invert == invert) {
 			/* found equivalent, report it */
 			return EEXIST;
 		}
 
-		xx_merge_filter(*pnext, pf_mask, invert);
-		return 0;
+		/* merge when possible */
+		if (xx_merge_filter(*pnext, pf_mask, rev, rev_match, invert))
+			return 0;
 	}
 
 	/* insert before */
@@ -1955,6 +2015,8 @@ static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int invert,
 		return ENOMEM;
 	n->cpuid = cpuid;
 	n->pfm = pf_mask;
+	n->rev = rev;
+	n->rev_match = rev_match;
 	n->invert = invert;
 	n->next = *pnext;
 	*pnext = n;
@@ -1982,34 +2044,51 @@ static void add_filter_list_to_list(struct microcode_filter_entry ** const base,
 		while (*pp && (*pp)->cpuid < e->cpuid)
 			pp = &((*pp)->next);
 
-		if (*pp && (*pp)->cpuid == e->cpuid) {
-			xx_merge_filter(*pp, e->pfm, e->invert);
-		} else {
+		if (!*pp || (*pp)->cpuid != e->cpuid ||
+		    !xx_merge_filter(*pp, e->pfm, e->rev, e->rev_match, e->invert)) {
 			/* insert before */
 			e->next = *pp;
 			*pp = e;
 			e = NULL;
 		}
-		if (e)
-			free(e);
+		free(e);
 	}
 }
 
+static int xx_compare_rev(const int32_t rev,
+			  const struct microcode_filter_entry *f)
+{
+	const enum iuc_rev_match_mode op = f->rev_match;
+
+	if ((op == IUCODE_REVFLT_ANY) ||
+	    (op == IUCODE_REVFLT_EQ && rev == f->rev) ||
+	    (op == IUCODE_REVFLT_LT && rev < f->rev)  ||
+	    (op == IUCODE_REVFLT_GT && rev > f->rev)) {
+		return !(f->invert);
+	}
+
+	return -1;
+}
+
 /* returns non-zero if selected */
 static int is_selected(const uint32_t cpuid,
 		       const uint32_t pf_mask,
+		       const int32_t rev,
 		       const struct microcode_filter_entry *f)
 {
+	int state = -1;
+
 	while (f && f->cpuid < cpuid)
 		f = f->next;
-	while (f && f->cpuid == cpuid) {
+	while (f && f->cpuid == cpuid && state < 0) {
 		if ((pf_mask & f->pfm) != 0 ||
-		    f->pfm == PFMASK_MATCH_ANY)
-			return !(f->invert);
+		    f->pfm == PFMASK_MATCH_ANY) {
+			state = xx_compare_rev(rev, f);
+		}
 		f = f->next;
 	}
 
-	return filter_list_allow;
+	return (state >= 0)? state : filter_list_allow;
 }
 
 /* Microcode extended signature deduplication */
@@ -2230,7 +2309,7 @@ static int xx_process_ucode_signature_cb(void * const userdata,
 			INTEL_UCLE_HASXST;
 		extsig_tables_in_use = 1;
 	}
-	if (is_selected(cpuid, pf_mask, uc_filter_list))
+	if (is_selected(cpuid, pf_mask, m.revision, uc_filter_list))
 		uce_flags |= (is_in_date_range(&m)) ?
 				INTEL_UCLE_SELECT : INTEL_UCLE_SELID;
 
@@ -2640,7 +2719,7 @@ static int xx_scan_handle_sig(const uint32_t id1, const uint32_t id2,
 		 * Get main processor signature from cpuid(1) EAX
 		 * and add it as a filter.
 		 */
-		switch (add_filter_to_list(sig, 0, 0, ucfp)) {
+		switch (add_filter_to_list(sig, 0, 0, IUCODE_REVFLT_ANY, 0, ucfp)) {
 		case ENOMEM:
 			print_err("out of memory");
 			return ENOMEM;
@@ -2656,8 +2735,22 @@ static int xx_scan_handle_sig(const uint32_t id1, const uint32_t id2,
 }
 
 #ifdef USE_CPUID_DEVICE
-static int check_cpuid_devs(__attribute__((unused)) const uint32_t sig,
-			    struct microcode_filter_entry ** const ucfp)
+static int is_base10_number(const char *s)
+{
+	if (!s || !*s)
+		return 0;
+
+	while(*s) {
+		if (*s < '0' || *s > '9')
+			return 0;
+		s++;
+	}
+
+	return 1;
+}
+
+/* requires a kernel driver, and *really* hurts on very large systems */
+static int xx_check_cpuid_devs(struct microcode_filter_entry ** const ucfp)
 {
 	uint32_t cpuid_buf[8];	/* two cpuid levels */
 	char cpuid_device[PATH_MAX];
@@ -2665,27 +2758,57 @@ static int check_cpuid_devs(__attribute__((unused)) const uint32_t sig,
 	int rc = 0;
 	unsigned int i = 0;
 	unsigned int ncpu = 0;
+	struct dirent *d;
+
+	DIR *cpudir = opendir(CPUID_DEVICE_PARENT);
+	if (!cpudir) {
+		int en = errno;
+		print_err("%s: could not open directory: %s", CPUID_DEVICE_PARENT, strerror(en));
+		return -1;
+	}
 
 	while (1) {
+		errno = 0;
+		d = readdir(cpudir);
+		if (!d) {
+			int err = errno;
+			if (unlikely(err)) {
+				print_err("%s: cannot walk directory: %s",
+					  CPUID_DEVICE_PARENT, strerror(err));
+				rc = -1; /* note that we skipped processors due to unexpected errors */
+			}
+			break; /* finish/abort walk */
+		}
+
+		/* Linux procfs supports d_type */
+		if (d->d_type != DT_DIR)
+			continue; /* next dentry */
+
+		/* must be [0-9]+, no trailling weirdness */
+		if (!is_base10_number(d->d_name))
+			continue; /* next dentry */
+
 		snprintf(cpuid_device, sizeof(cpuid_device),
-			 CPUID_DEVICE_BASE, i);
-		cpuid_fd = open(cpuid_device, O_RDONLY);
+			 CPUID_DEVICE_BASE, d->d_name);
+
+		errno = EINTR;
+		cpuid_fd = -1;
+		while (cpuid_fd == -1 && errno == EINTR)
+			cpuid_fd = openat(dirfd(cpudir), cpuid_device, O_RDONLY | O_CLOEXEC);
 		if (cpuid_fd == -1) {
 			int en = errno;
 
-			print_msg(4, "%s: returned error status on open(): %s",
-				  cpuid_device, strerror(en));
+			print_msg(4, "%s/%s: returned error status on open(): %s",
+				  CPUID_DEVICE_PARENT, cpuid_device, strerror(en));
 
-			if (en == EINTR)
-				continue; /* try again */
 			if (en == ENOENT)
 				break;    /* cpuid device inode not found */
 			if (en == ENXIO || en == EIO) {
 				/* Linux cpuid driver: ENXIO: offline; EIO: no cpuid support */
 				print_msg(2, "processor %u is offline or has no cpuid support", i);
 			} else {
-				print_msg(2, "%s: cannot open cpuid device node: %s",
-					cpuid_device, strerror(en));
+				print_msg(2, "%s/%s: cannot open cpuid device node: %s",
+					CPUID_DEVICE_PARENT, cpuid_device, strerror(en));
 				rc = -1; /* note that we skipped processors due to unexpected errors */
 			}
 
@@ -2694,13 +2817,14 @@ static int check_cpuid_devs(__attribute__((unused)) const uint32_t sig,
 			continue;
 		}
 
-		print_msg(3, "trying to get CPUID information from %s",
-			  cpuid_device);
+		print_msg(3, "trying to get CPUID information from %s/%s",
+			  CPUID_DEVICE_PARENT, cpuid_device);
 		if (read(cpuid_fd, &cpuid_buf, sizeof(cpuid_buf)) == -1) {
-			print_err("%s: access to CPUID(0) and CPUID(1) failed: %s",
-				  cpuid_device, strerror(errno));
+			print_err("%s/%s: access to CPUID(0) and CPUID(1) failed: %s",
+				  CPUID_DEVICE_PARENT, cpuid_device, strerror(errno));
 			/* this is in the should not happen list, so abort */
 			close(cpuid_fd);
+			closedir(cpudir);
 			return 1;
 		}
 
@@ -2716,6 +2840,8 @@ static int check_cpuid_devs(__attribute__((unused)) const uint32_t sig,
 		i++;
 	};
 
+	closedir(cpudir);
+
 	if (i == 0 && ncpu == 0) {
 		print_err("cpuid kernel driver unavailable");
 		return -1;
@@ -2732,16 +2858,22 @@ static int check_cpuid_devs(__attribute__((unused)) const uint32_t sig,
 	return rc;
 }
 #else
+static int xx_check_cpuid_devs(__attribute__((unused)) struct microcode_filter_entry ** const ucfp)
+{
+	print_msg(1, "support for exact system scan disabled at compile time");
+	return -1;
+}
+#endif /* USE_CPUID_DEVICE */
 
-/* this hurts a lot less on very big systems... */
-static int check_cpuid_devs(uint32_t sig, struct microcode_filter_entry ** const ucfp)
+
+/* xx_add_all_steppings(cpuid) hurts a lot less on very big systems... */
+static int xx_add_all_steppings(uint32_t sig, struct microcode_filter_entry ** const ucfp)
 {
 	unsigned int i;
 
-	print_msg(2, "assuming all processors have the same type, family and model");
 	sig |= 0xf;
 	for (i = 0; i < 0x10; i++) {
-		if (add_filter_to_list(sig, 0, 0, ucfp) == ENOMEM) {
+		if (add_filter_to_list(sig, 0, 0, IUCODE_REVFLT_ANY, 0, ucfp) == ENOMEM) {
 			print_err("out of memory");
 			return ENOMEM;
 		}
@@ -2749,14 +2881,46 @@ static int check_cpuid_devs(uint32_t sig, struct microcode_filter_entry ** const
 	}
 	return 0;
 }
-#endif
 
-static int scan_system_processors(void)
+/* Handle mixed-signature systems, even if Intel still hasn't admited that
+ * they will exist in the SDM, the writing is already out in the wall */
+static int xx_add_all_steppings_for_every_sig(uint32_t sig, struct microcode_filter_entry ** const ucfp)
+{
+	struct microcode_filter_entry *fl = NULL;
+	struct microcode_filter_entry *p;
+	int rc;
+
+	/* failsafe of the failsafe: add running processor sig */
+	rc = xx_add_all_steppings(sig, &fl);
+
+	/* creating a new list is just plain safer in the long run */
+	p = *ucfp;
+	while (p && !rc) {
+		/* handle mixed sig systems, unlikely as that might be */
+		if (p->cpuid != sig)
+			rc = xx_add_all_steppings(p->cpuid, &fl);
+		p = p->next;
+	}
+
+	if (!rc) {
+		free_filter_list(*ucfp);
+		*ucfp = fl;
+	} else {
+		free_filter_list(fl);
+	}
+
+	return rc;
+}
+
+static int scan_system_processors(unsigned int strategy,
+	                          struct microcode_filter_entry ** const filter_list)
 {
 	uint32_t id0, id1, id2, id3, sig, idx;
 	struct microcode_filter_entry *uc_cpu = NULL;
 	int rc = 0;
 
+	assert(filter_list);
+
 	print_msg(3, "trying to get CPUID information directly");
 	if (!(__get_cpuid(0, &id0, &id1, &id2, &id3) &&
 	      __get_cpuid(1, &sig, &idx, &idx, &idx))) {
@@ -2775,7 +2939,15 @@ static int scan_system_processors(void)
 
 	switch (xx_scan_handle_sig(id1, id2, id3, sig, &uc_cpu)) {
 	case 0:
-		rc = check_cpuid_devs(sig, &uc_cpu);
+		if (strategy == 2) {
+			if (xx_check_cpuid_devs(&uc_cpu)) {
+				print_warn("exact cpuid signature scan failed, switching to failsafe strategy");
+				rc = xx_add_all_steppings_for_every_sig(sig, &uc_cpu);
+			}
+		} else {
+			print_msg(2, "assuming all processors have the same type, family and model");
+			rc = xx_add_all_steppings(sig, &uc_cpu);
+		}
 		break;
 	case ENXIO:
 		print_msg(1, "running on a non-Intel processor");
@@ -2792,18 +2964,81 @@ static int scan_system_processors(void)
 
 	if (uc_cpu) {
 		/* tie linked lists */
-		add_filter_list_to_list(&uc_filter_list, uc_cpu);
+		add_filter_list_to_list(filter_list, uc_cpu);
 		uc_cpu = NULL;
 	}
 
 	return (rc > 0) ? rc : 0;
 }
 
+static int cmdline_queue_ucode_filter(uint32_t cpuid, uint32_t pf_mask,
+				      int32_t rev, enum iuc_rev_match_mode rev_match,
+				      int invert)
+{
+	struct microcode_filter_entry *n;
+
+	n = malloc(sizeof(struct microcode_filter_entry));
+	if (!n)
+		return ENOMEM;
+	n->cpuid = cpuid;
+	n->pfm = pf_mask;
+	n->rev = rev;
+	n->rev_match = rev_match;
+	n->invert = invert;
+	n->next = NULL;
+
+	if (uc_filter_queue_tail) {
+		uc_filter_queue_tail->next = n;
+	} else {
+		uc_filter_queue = n;
+	}
+
+	uc_filter_queue_tail = n;
+
+	return 0;
+}
+
+static int process_ucode_filter_queue(void)
+{
+	struct microcode_filter_entry *p = uc_filter_queue;
+	int rc = 0;
+
+	while(p && !rc) {
+		if (p->cpuid != IUCODE_FILTERMASK_SCANCPUS) {
+			rc = add_filter_to_list(p->cpuid, p->pfm, p->rev, p->rev_match, p->invert, &uc_filter_list);
+			if (rc == EEXIST)
+				rc = 0;
+		} else {
+			rc = scan_system_processors(p->pfm, &uc_filter_list);
+		}
+
+		p = p->next;
+	}
+
+	uc_filter_queue_tail = NULL;
+	free_filter_list(uc_filter_queue);
+	uc_filter_queue = NULL;
+
+	switch(rc) {
+	case 0:
+		return 0;
+
+	case ENOMEM:
+		print_err("Cannot add filter entry: out of memory");
+		break;
+	case EINVAL:
+		print_err("Internal error while processing filter list");
+		break;
+	}
+
+	return 1;
+}
+
 /* Command line processing */
 
 static const char program_version[] =
 	PROGNAME " " VERSION "\n"
-	"Copyright (c) 2010-2017 by Henrique de Moraes Holschuh\n\n"
+	"Copyright (c) 2010-2018 by Henrique de Moraes Holschuh\n\n"
 
 	"Based on code from the Linux microcode_intel driver and from\n"
 	"the microcode.ctl package, copyright (c) 2000 by Simon Trimmer\n"
@@ -2851,6 +3086,7 @@ enum {
 	IUCODE_ARGP_DATEAFTER,
 	IUCODE_ARGP_DATEFSTRICT,
 	IUCODE_ARGP_DATEFLOOSE,
+	IUCODE_ARGP_SCANSYSTEMOPT,
 	IUCODE_ARGP_EIRFS,
 	IUCODE_ARGP_MINSIZE_EIRFS,
 	IUCODE_ARGP_DFLSIZE_EIRFS,
@@ -2870,19 +3106,29 @@ static const struct argp_option cmdline_options[] = {
 	   "selected by filename suffix)",
 	  10 },
 
-	{ NULL, 's', "! | [!]signature[,pf_mask]", 0,
-	   "Select microcodes by the specified signature and processor "
-	   "flags mask (pf_mask).  Specify more than once to select/unselect "
-	   "more microcodes.  Prefix with ! to unselect microcodes.  "
+	{ NULL, 's', "! | [!]signature[,[pf_mask][,[lt:|eq:|gt:]revision]]", 0,
+	   "Select microcodes by the specified signature, processor "
+	   "flags mask (pf_mask), and revision.  Optionally, prefix revision "
+	   "with eq: (equal -- implied there is no prefix), lt: (less than) "
+	   "or gt: (greater than).  "
+	   "Specify more than once to select/unselect more microcodes.  "
+	   "Prefix with ! to unselect microcodes.  "
 	   "Use -s ! to disable the default behavior of selecting all "
 	   "microcodes when no -s or -S filter is specified",
 	  20 },
-	{ "scan-system", 'S', NULL, 0,
+	{ NULL, 'S', NULL, 0,
+	   "Same as --scan-system=auto",
+	  21 },
+	{ "scan-system", IUCODE_ARGP_SCANSYSTEMOPT, "mode", OPTION_ARG_OPTIONAL,
 	   "Select microcodes based on the running system processor(s).  "
 	   "Can be combined with the -s option, and can be used only once.  "
 	   "Microcodes selected by --scan-system can be unselected by a "
-	   "later -s !<signature> option",
-	  20 },
+	   "later -s !<signature> option.  The optional mode argument "
+	   "selects the strategy: 0 or auto (default); 1 or fast (good for "
+	   "most systems, including mixed-stepping); and 2 or exact (slow, "
+	   "supports multi-signature systems, requires the cpuid kernel "
+	   "driver and might require root access)",
+	  22 },
 
 	{ "downgrade", IUCODE_ARGP_DOWNGRADE, NULL, 0,
 	   "Instead of discarding microcodes based on revision level, "
@@ -2976,6 +3222,8 @@ static const struct argp_option cmdline_options[] = {
 };
 static const char cmdline_nonarg_doc[] = "[[-t<type>] filename] ...";
 
+static const char * const cmdline_scan_system_tbl[] = { "auto", "fast", "exact", NULL };
+
 static int new_filename(const char * const fn,
 			const intel_ucode_file_type_t ftype)
 {
@@ -3022,15 +3270,17 @@ static void free_filename_list(void)
 	input_files = NULL;
 }
 
-/* -s ! | [!]cpuid[,pf_mask] */
-static int add_ucode_filter(const char *arg)
+/* -s ! | [!]cpuid[,[pf_mask][,[<revmatch_operator>]rev]] */
+static int cmdline_parse_ucode_filter(const char *arg)
 {
 	char *p;
 	uint32_t acpuid, amask;
-	int invert;
+	int32_t arev;
+	int invert, arev_match;
 
 	amask = 0;
 	invert = 0;
+	arev = 0;
 
 	while (isspace(*arg))
 		arg++;
@@ -3051,28 +3301,65 @@ static int add_ucode_filter(const char *arg)
 		}
 	}
 
+	/* cpuid, mandatory */
 	if (!*arg || parse_u32(arg, &p, 0, &acpuid))
 		return EINVAL;
 	while (isspace(*p))
 		p++;
 
+	/* pf_mask, optional */
 	if (*p == ',') {
 		p++;
+		if (*p != ',') {
+			arg = p;
+			if (!*arg || parse_u32(arg, &p, 0, &amask))
+				return EINVAL;
+		} /* else amask = MATCH_ANY */
+		while (isspace(*p))
+			p++;
+	} else if (*p) {
+		return EINVAL;
+	}
+
+	/* rev, optional */
+	if (*p == ',') {
+		p++;
+		while (isspace(*p))
+			p++;
+		if (!*p)
+			return EINVAL;
+
+		/* revison match operator, optional */
+		int i = 0;
+		while (i < IUCODE_REVFLT_SIZE &&
+		       (!iuc_rev_match_mode_s[i] ||
+		        strncasecmp(p, iuc_rev_match_mode_s[i],
+			            strlen(iuc_rev_match_mode_s[i]))))
+			i++;
+		if (i < IUCODE_REVFLT_SIZE) {
+			arev_match = i;
+			p += strlen(iuc_rev_match_mode_s[i]);
+		} else {
+			arev_match = IUCODE_REVFLT_EQ;
+		}
+
+		/* revision */
 		arg = p;
-		if (!*arg || parse_u32(arg, &p, 0, &amask))
+		if (!*arg || parse_s32e(arg, &p, 0, &arev))
 			return EINVAL;
 		while (isspace(*p))
 			p++;
-		if (*p)
-			return EINVAL;
-	} else if (*p) {
+	} else {
+		arev_match = IUCODE_REVFLT_ANY;
+	}
+	if (*p) {
 		return EINVAL;
 	}
 
 	if (!invert)
 		filter_list_allow = 0;
 
-	return add_filter_to_list(acpuid, amask, invert, &uc_filter_list);
+	return cmdline_queue_ucode_filter(acpuid, amask, arev, arev_match, invert);
 }
 
 /* YYYY-MM-DD */
@@ -3114,6 +3401,55 @@ static int cmdline_get_date(const char *arg, uint32_t *d)
 	return 0;
 }
 
+static int cmdline_get_int(const char *arg, int pmin, int pmax, int *d)
+{
+	char *p;
+	long int l;
+
+	if (!arg)
+		return 0;
+
+	errno = 0;
+	l = strtol(arg, &p, 10);
+	if (errno || p == arg)
+		return errno ? errno : EINVAL;
+	if (l > INT32_MAX || l < INT32_MIN)
+		return ERANGE;
+	if (l < pmin || l > pmax)
+		return EINVAL;
+
+	while (isspace(*p))
+		p++;
+	if (*p)
+		return EINVAL;
+
+	if (d)
+		*d = (int) l;
+
+	return 0;
+}
+
+/* note: does not tolerate trailing blanks */
+static int cmdline_get_enumstr(const char *arg, const char * const table[],
+			       int *d)
+{
+	int r = 0;
+
+	if (!arg || !table)
+		return 0;
+
+	while (table[r]) {
+		if (!strcasecmp(arg, table[r])) {
+			if (d)
+				*d = r;
+			return 0;
+		}
+		r++;
+	}
+
+	return ENOENT;
+}
+
 static error_t cmdline_do_parse_arg(int key, char *arg,
 				    struct argp_state *state)
 {
@@ -3237,29 +3573,61 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 		break;
 
 	case 's':
-		rc = add_ucode_filter(arg);
+		rc = cmdline_parse_ucode_filter(arg);
 		switch (rc) {
 		case 0:
-		case EEXIST:
 			break; /* success */
 		case EINVAL:
 			argp_error(state, "invalid filter: '%s'", arg);
 			break; /* not reached */
 		default:
 			argp_failure(state, EXIT_SWFAILURE, rc,
-				     "could not add filter '%s'", arg);
+				     "could not queue filter '%s'", arg);
 		}
 		command_line_actions |= IUCODE_F_UCSELECT;
 		break;
+
 	case 'S':
+		/*
+		 * -S and --scan-system cannot be handled the same way
+		 * because -S cannot be made OPTION_ARG_OPTIONAL: it would
+		 * break command-line backwards compatibility when people
+		 * specify several short options together and -S is not the
+		 * last option in the chain
+		 */
+		if (command_line_actions & IUCODE_DO_SELPROC)
+			argp_error(state,
+				   "--scan-system option can be specified only once");
+		/*
+		 * note: the pfm field of the queue filter item is used to
+		 * encode the scan-system strategy.
+		 */
+		rc = cmdline_queue_ucode_filter(IUCODE_FILTERMASK_SCANCPUS, 0, 0, 0, 0);
+		if (rc)
+			argp_failure(state, EXIT_SWFAILURE, rc,
+				     "could not queue --scan-system action");
+		command_line_actions |= IUCODE_DO_SELPROC | IUCODE_F_UCSELECT;
+		break;
+	case IUCODE_ARGP_SCANSYSTEMOPT:
 		if (command_line_actions & IUCODE_DO_SELPROC)
 			argp_error(state,
 				   "--scan-system option can be specified only once");
 
-		if (scan_system_processors()) {
-			argp_failure(state, EXIT_SWFAILURE, 0,
-				"fatal failure while scanning for system processors");
-		}
+		int scan_system_strategy = 0;
+		if (cmdline_get_int(arg, 0, 2, &scan_system_strategy) &&
+		    cmdline_get_enumstr(arg, cmdline_scan_system_tbl, &scan_system_strategy))
+			argp_error(state, "invalid --scan-system mode: '%s'", arg);
+
+		/*
+		 * note: the pfm field of the queue filter item is used to
+		 * encode the scan-system strategy.
+		 */
+		rc = cmdline_queue_ucode_filter(IUCODE_FILTERMASK_SCANCPUS,
+		                                (uint32_t)scan_system_strategy,
+						0, 0, 0);
+		if (rc)
+			argp_failure(state, EXIT_SWFAILURE, rc,
+				     "could not queue --scan-system action");
 		command_line_actions |= IUCODE_DO_SELPROC | IUCODE_F_UCSELECT;
 		break;
 
@@ -3335,6 +3703,11 @@ int main(int argc, char *argv[])
 		goto do_nothing;
 	}
 
+	if (process_ucode_filter_queue()) {
+		rc = EXIT_SWFAILURE;
+		goto err_exit;
+	}
+
 	if (command_line_actions & IUCODE_DO_LOADFILE) {
 		struct filename_list *fn = input_files;
 		while (fn && next_bundle_id) {
